// Serialized ImGui windows docking configuration
// generated using <geogram_program> gui:expert=true
// then Windows->Export gui state to C++
const char gui_state[] = u8"[Window][DockSpace]    Pos=0,0    Size=1024,1024    Collapsed=0        [Window][ Object]    Pos=60,60    Size=172,458    Collapsed=0        [Docking][Data]    DockSpace         ID=0x09EF459F Pos=0,24 Size=1024,999 Split=Y      DockNode        ID=0x00000001 Parent=0x09EF459F SizeRef=1024,825 Split=X        DockNode      ID=0x00000005 Parent=0x00000001 SizeRef=158,764 Split=Y SelectedTab=0x07957BC8          DockNode    ID=0x00000007 Parent=0x00000005 SizeRef=158,470 Split=Y SelectedTab=0x07957BC8            DockNode  ID=0x00000009 Parent=0x00000007 SizeRef=158,467 HiddenTabBar=1 SelectedTab=0x07957BC8            DockNode  ID=0x0000000A Parent=0x00000007 SizeRef=158,295 HiddenTabBar=1 SelectedTab=0x4177D348          DockNode    ID=0x00000008 Parent=0x00000005 SizeRef=158,292 HiddenTabBar=1 SelectedTab=0xBD088ACE        DockNode      ID=0x00000006 Parent=0x00000001 SizeRef=696,764 CentralNode=1 HiddenTabBar=1 SelectedTab=0x080FC883      DockNode        ID=0x00000002 Parent=0x09EF459F SizeRef=1024,172 HiddenTabBar=1 SelectedTab=0xF9BEF62A        ";
