-- Dessin d'un "creeper"
-- Par Nathan Levy, Decembre 2016

require("pixel")

function GLUP.draw_scene()
   GLUP.SetCellsShrink(0.1)
   GLUP.Enable(GLUP.DRAW_MESH)
   GLUP.Enable(GLUP.VERTEX_COLORS)
   pixBegin();
   col("green")
   pix(1,1)
   pix(2,1)
   pix(3,1)
   pix(4,1)
   pix(5,1)
   pix(6,1)
   pix(7,1)
   pix(8,1)
   pix(1,2)
   col("black")
   pix(2,2)
   pix(3,2)
   col("green")
   pix(4,2)
   pix(5,2)
   col("black")
   pix(6,2)
   pix(7,2)
   col("green")
   pix(8,2)
   pix(1,3)
   col("black")
   for x=1,7,1 do
      pix(x,3)
   end
   col("green")
   pix(8,3)
   pix(1,4)
   pix(2,4)
   pix(3,4)
   col("black")
   pix(4,4)
   pix(5,4)
   col("green")
   pix(6,4)
   pix(7,4)
   pix(8,4)
   for x=1,8,1 do
      pix(x,5)
   end
   pix(1,6)
   col("black")
   pix(2,6)
   pix(3,6)
   col("green")
   pix(4,6)
   pix(5,6)
   col("black")
   pix(6,6)
   pix(7,6)
   col("green")
   pix(8,6)
   pix(1,7)
   col("black")
   pix(2,7)
   pix(3,7)
   col("green")
   pix(4,7)
   pix(5,7)
   col("black")
   pix(6,7)
   pix(7,7)
   col("green")
   pix(8,7)
   for x=1,8,1 do
      pix(x,8)
   end
   
   pixEnd()
end
